//uniform... von OpenGL...
uniform sampler1D sTransfer;
uniform sampler1D sCartoon;
uniform sampler2D sBackFace;
uniform sampler3D volume_tex;
uniform sampler3D sPartiel;
uniform vec3 sSliding;
uniform int IsMode;
uniform float sSlideValue;
uniform int adaptiv;
uniform int sMaxIntensity;
uniform int sPreshading;

//varying... von Vertex Shader
varying vec4 hposition;
varying vec4 position;
varying mat4 sAdaptivPlane;

//globlae variable
bool preprocessing = true; //Preprocessing wohl ned ganz ok?

vec3 readTexture(vec3 pos)
{
	if(pos.x > 1.0) pos.x = 1.0;
	if(pos.y > 1.0) pos.y = 1.0;
	if(pos.z > 1.0) pos.z = 1.0;
	if(pos.x < -1.0) pos.x = -1.0;
	if(pos.y < -1.0) pos.y = -1.0;
	if(pos.z < -1.0) pos.z = -1.0;
	
	return pos;
}

//Phong shading .... siehe Paper (Capter 3)
vec3 phong(vec3 lightcolor, float distance, float kamb, float kdiff, float kspec, float n, float k1, float k2, vec3 norm, vec3 l, vec3 h, float shadow)
{
	vec3 normal = normalize(norm);
	return	(lightcolor*kamb)+	(lightcolor/(k1+(k2*distance))*(kdiff*min(1.0-shadow,max(dot(normal, l),0.0))+	(kspec*min(1.0-shadow,pow(max(dot(normal, h),0.0),n)))));
	//return norm;
}

float innerProduct(vec3 v1, vec3 v2) {

	return (v1.x * v2.x) + (v1.y * v2.y) + (v1.z * v2.z);
}

vec3 calcNormal(vec3 vec, bool preprocessing)
{
	vec3 norm2;
	
	//Partiellen Ableitungen berechne ("on the fly") oder lade (preprocessing)
	if(adaptiv == 0)	
	{
		vec4 norm3 = texture3D(sPartiel, vec);	//Preprocessing
		norm2.rgb = norm3.rgb;
		//normalize(norm2);
	}
	else{
		norm2 = vec3( 
					texture3D(volume_tex, vec.xyz+vec3(0.01,0.00,0.00)).r-texture3D(volume_tex, vec.xyz-vec3(0.01,0.00,0.00)).r,
					texture3D(volume_tex, vec.xyz+vec3(0.00,0.01,0.00)).r-texture3D(volume_tex, vec.xyz-vec3(0.00,0.01,0.00)).r,
					texture3D(volume_tex, vec.xyz+vec3(0.00,0.00,0.01)).r-texture3D(volume_tex, vec.xyz-vec3(0.00,0.00,0.01)).r
					);	//"on the fly"
		norm2*= 1.0/2.0;
	}
	
	return norm2;
}

vec3 cellshading(vec3 lightcolor, float distance, float kamb, float kdiff, float kspec, float n, float k1, float k2, vec3 norm, vec3 l, vec3 h,vec3 view, vec3 vec)
{
	vec3 normal = normalize(norm);
	l = normalize(l);
       
	/* 
    //vec3 norm3 = calcNormal(vec+vec3(0,0,0.001), preprocessing);
    float edgefactor = clamp(length(norm)*kdiff*max(dot(normal, l),0.0),0.0,1.0);
    if(edgefactor >= 0.05)	edgefactor = 1.0;
    else					edgefactor = 0.0;
    */
	
	vec3 color = texture1D(sCartoon, max(dot(normal, l),0.0)).xyz;
	color =	(color*kamb)+	(color/(k1+(k2*distance))*(kdiff*max(dot(normal, l),0.0)+(kspec*pow(max(dot(normal, h),0.0),n))));
	
	float edgefactor = clamp(length(norm)*kdiff*max(dot(normal, l),0.0),0.0,1.0);
    if(edgefactor < 0.01)	color = vec3(0.0);
    
	return	color;
}

float getShadow(vec3 lpos, vec3 vec, float stepsize )
{
	vec3 direction = lpos - vec;					//Direction vom Strahl
	float len = length(direction);					//Lnge vom Strahl
	vec3 norm_dir = normalize(direction);			//Directionvektor normalisieren

	vec3 delta_dir = direction * stepsize;			
	float delta_dir_len = length(delta_dir);
	float length_acc = 0;
	float alpha_acc = 0;
	for(int i = 0; i < 100 && (length_acc <= len);i++)
	{			
		vec4 lookup = texture3D(sPartiel, vec);
		vec3 offset = normalize(lookup.xyz)*stepsize;
		float sval = texture3D(sPartiel, vec + offset).a;
		vec4 transfer = texture1D(sTransfer, sval);
		float alpha_sample = transfer.a;	
				
		vec4 color_sample = transfer;
						
		if(alpha_sample > 0.2)
		{
			alpha_acc = 0.9;
			return alpha_acc;
		}
			
		alpha_acc += (1.0-alpha_acc)*(alpha_sample/50.0f);
			
		vec += delta_dir;
		length_acc+=len;
	}
	
	return alpha_acc;
}

//Main Methode
void main()
{ 
	if(IsMode == 1)
	{
		float stepsize = 1.0/200.0;						//Definiert die Schrittweite des Strahles
		float stepsize2 = 1.0/1000.0;						//Definiert die Schrittweite des Strahles
		vec3 bposition = position.xyz / position.w;		//Erzeugt Texturkoordinaten		
		
		vec3 back = texture2D(sBackFace, ((bposition.xy+1.0)/2.0).st).xyz;
		vec3 start = hposition.xyz;
	
		//FRONT TO BACK
		vec3 direction = back - start;					//Direction vom Strahl
		vec3 vec = start;
		
		//BACK TO FRONT
		//vec3 direction = start-back;					//Direction vom Strahl
		//vec3 vec = back;
		
		float len = length(direction);					//Lnge vom Strahl
		vec3 norm_dir = normalize(direction);			//Directionvektor normalisieren

		vec3 delta_dir = direction * stepsize;			
		float delta_dir_len = length(delta_dir);
		
		float alpha_acc = 0.0;
		vec4 color_sample;
		float alpha_sample = 0.0;
		float length_acc = 0.0;
		vec4 col_acc = vec4(0,0,0,0);
		
		vec3 viewingdirection = normalize(back - start);
	
		vec3 lpos = vec3(0.0,0.0,1.0);		
		vec4 trans = vec4(0,0,0,0);
		
		for(int i = 0; i < 200 && (length_acc <= len);i++)
		{
			vec3 norm2 = calcNormal(vec, preprocessing);			
			float lumin	= 0.0f;				
			
			float sval = texture3D(sPartiel, vec).a;
			vec4 transfer = texture1D(sTransfer, sval);
			alpha_sample = transfer.a;	
			
			vec4 color_sample = transfer;
			vec3 l =  lpos - vec;// light vector
			vec3 h = normalize(norm_dir + l); // half vector  		
			
			if(sPreshading == 0) 
			{		
				color_sample = vec4(phong(color_sample.xyz, length(vec), 0.01, 0.5, 0.8, 1.0, 0.3, 0.1, norm2, normalize(l), h, getShadow(lpos, vec, 1/100 )),1.0); 
			}
			else
			{
				color_sample = vec4(phong(color_sample.xyz, length(vec), 0.01, 0.5, 0.8, 1.0, 0.3, 0.1, norm2, normalize(l), h, 0.0),1.0); 
			}
   
			
			if(sMaxIntensity == 1)
			{				
				//MIP
				if(alpha_sample > 0.05)
				{
					col_acc   = color_sample;
					alpha_acc = alpha_sample;
					break;
				}
			}
			else	
			{	
				//FRONT TO BACK		
				col_acc   += (1.0-alpha_acc) * color_sample * alpha_sample;
				alpha_acc += (1.0-alpha_acc)*alpha_sample;	
			}
		
			vec += delta_dir;
			length_acc += delta_dir_len;
		}
		
		vec3 oldvec = vec;
		
		
		gl_FragColor = col_acc;
		
	}
	else
	{
		vec3 bposition = position.xyz / position.w;		//Erzeugt Texturkoordinaten
		vec3 tex;
		if(IsMode == 2)	tex = vec3( ((bposition.xy+1.0)/2.0).x,sSlideValue,((bposition.xy+1.0)/2.0).y);
		if(IsMode == 3)	tex = vec3( sSlideValue,((bposition.xy+1.0)/2.0).xy);
		if(IsMode == 4)	tex = vec3( ((bposition.xy+1.0)/2.0).xy,sSlideValue);
		
		//gl_FragColor = texture1D(sTransfer, texture3D(volume_tex, tex).r);
		gl_FragColor = texture1D(sTransfer, texture3D(sPartiel, tex).a);
		//gl_FragColor /= gl_FragColor.w;
	}
}