//uniform... von OpenGL...
uniform sampler1D sTransfer;
uniform sampler2D sBackFace;
uniform sampler2D sBackground;
uniform sampler3D sPartiel;
uniform vec3 sLightSourcePos;
uniform int IsMode;
uniform float sSlideValue;
uniform int preprocess;
uniform int sMaxIntensity;
uniform int sPreshading;
uniform int sInterpolation;
uniform float stepsize;
uniform vec3 resolution;
uniform int sAdaptiv;
uniform int sClipping;
uniform int sTransferGradient;

//varying... von Vertex Shader
varying vec4 hposition;
varying vec4 position;
varying vec3 norClipping;
varying float d;


//Phong shading .... siehe Paper (Capter 3)
float phong(vec3 lightcolor, float distance, float kamb, float kdiff, float kspec, float n, float k1, float k2, vec3 norm, vec3 l, vec3 h)
{
	vec3 normal = normalize(norm);
	return	(kamb)+	(1.0/(k1+(k2*distance))*(kdiff*min(1.0,max(dot(normal, l),0.0))+	(kspec*min(1.0,pow(max(dot(normal, h),0.0),n)))));
}


vec3 calcNormal(vec3 vec)
{
	vec3 norm2;
	
	//Partiellen Ableitungen berechne ("on the fly") oder lade (preprocessing)
	if(preprocess == 0)	
	{
		norm2 = texture3D(sPartiel, vec).xyz;	//Preprocessing
	}
	else{
		norm2 = vec3( 
					texture3D(sPartiel, vec.xyz+vec3(resolution.x,0.00,0.00)).a-texture3D(sPartiel, vec.xyz-vec3(resolution.x,0.00,0.00)).a,
					texture3D(sPartiel, vec.xyz+vec3(0.00,resolution.y,0.00)).a-texture3D(sPartiel, vec.xyz-vec3(0.00,resolution.y,0.00)).a,
					texture3D(sPartiel, vec.xyz+vec3(0.00,0.00,resolution.z)).a-texture3D(sPartiel, vec.xyz-vec3(0.00,0.00,resolution.z)).a
					);	//"on the fly" keine Ahnung was schner ist
	}
	
	return norm2;
}


//Main Methode
void main()
{ 
	float stepsize2 = stepsize*0.5;						//Definiert die Schrittweite des Strahles
	vec3 bposition = position.xyz / position.w;		//Erzeugt Texturkoordinaten		
	vec3 back = texture2D(sBackFace, ((bposition .xy+1.0)/2.0).st).xyz;
	vec3 start = hposition.xyz;
	
	//FRONT TO BACK
	vec3 direction = back - start;					//Direction vom Strahl
	vec3 vec = start;
			
	float len = length(direction);					//Lnge vom Strahl
	vec3 norm_dir = normalize(direction);			//Directionvektor normalisieren
	vec3 delta_dir = direction * stepsize;			
	float delta_dir_len = length(delta_dir);
		
	float alpha_acc = 0.0;
	vec4 color_sample;
	float alpha_sample = 0.0;
	float length_acc = 0.0;
	vec4 col_acc = vec4(0,0,0,0);
		
	vec3 viewingdirection = normalize(back - start);
	vec3 greatest= start;
	float shadingintensity = 0.0;
	float maxshadingintensity = 0.0;
		
	for(int i = 0;  i < 10000 ;i++)
	{
		if((alpha_acc >= 1.0)||(length_acc > len)) break;
			
		vec3 norm2;
		float sval;
		
		if(sInterpolation == 1)
		{
			norm2 = calcNormal(vec);			
			sval = texture3D(sPartiel, vec).a;
		
			vec4 a1 = (0.125*texture3D(sPartiel, vec.xyz+vec3(resolution.x,0.00,0.00)))
						+(0.125*texture3D(sPartiel, vec.xyz-vec3(resolution.x,0.00,0.00)))
				         +(0.125*texture3D(sPartiel, vec.xyz+vec3(0.00,0.00,resolution.z)))
						 +(0.125*texture3D(sPartiel, vec.xyz-vec3(0.00,0.00,resolution.z)))
						 +(0.125*texture3D(sPartiel, vec.xyz+vec3(resolution.x,0.00,resolution.z)))
						 +(0.125*texture3D(sPartiel, vec.xyz-vec3(resolution.x,0.00,resolution.z)))
				 		 +(0.125*texture3D(sPartiel, vec.xyz+vec3(-resolution.x,0.00,resolution.z)))
			 			 +(0.125*texture3D(sPartiel, vec.xyz-vec3(-resolution.x,0.00,resolution.z)));
		
			sval = (a1.a * 0.5)+(sval*0.5);
			norm2 = (norm2.xyz * 0.5)+(a1.xyz*0.5);
		}
		else
		{
			norm2 = calcNormal(vec);			
			sval = texture3D(sPartiel, vec).a;
		}	
		
		vec3 l =  vec-sLightSourcePos;// light vector
		vec3 h = normalize(norm_dir + l); // half vector  		
	
		vec4 transfer = texture1D(sTransfer, sval);
		shadingintensity = phong(color_sample.xyz, length(l), 0.01, 0.5, 0.8, 0.8, 0.3, 0.1, normalize(norm2), normalize(l), h);
		alpha_sample = transfer.a * pow(length(norm2),pow(1.0*shadingintensity*(1.0-(length(start-vec)/10.0))*(1.0-alpha_sample),0.8));
		
		if (((d + dot(norClipping.xyz,vec.xyz)) < 0.0) || (sClipping == 0))
		{
			//MIP
			if(alpha_sample > alpha_acc )
			{
				alpha_acc = alpha_sample;
				greatest = vec;
				maxshadingintensity = shadingintensity;
			}
		}
	
		if(sAdaptiv == 1)
		{
			vec3 deltad = (direction * (stepsize*(1.0-sval))) + (direction * (stepsize2*sval));
				vec +=deltad;
				length_acc += abs(length(deltad));
		}
		else
		{
			vec += delta_dir;
			length_acc += delta_dir_len;
		}
	}
		
	float sval = texture3D(sPartiel, greatest).a;			
	vec4 transfer = texture1D(sTransfer, sval);
		
	if(sPreshading == 0) 
	{		
		col_acc = vec4(transfer.xyz*maxshadingintensity,1.0); 
	}
	else col_acc = transfer;
	col_acc.a = alpha_acc;
		
	if(sPreshading == 0)	gl_FragColor = mix(col_acc,texture2D(sBackground, ((bposition.xy+1.0)/2.0).st),1.0-alpha_acc);
	else					gl_FragColor = col_acc;
}